import machine
import neopixel
import os
import utime
import _thread


NUM_TURN_LEDS_3 = 25
NUM_TURN_LEDS_2 = 24
NUM_TURN_LEDS_1 = 26

NUM_FOG_LEDS = 39  # 11.4 watts 2.28 amps @ 5V, voltage drop: 18AWG 3 conductors 0.097V @ 10 feet
NUM_MARKER_LEDS = 13  # 7.2 watts 1.44 amps @ 5V, voltage drop: 18AWG 3 conductors 0.061V @ 10 feet

#
#
# NUM_TURN_LEDS = 72  # 21.6 watts 4.32 amps @ 5V, voltage drop: 18AWG 3 consuctors 0.18V @ 10 feet
# NUM_FOG_LIGHT_LEDS = 45  # 13.5 watts 2.7 amps @ 5V, voltage drop: 18AWG 3 conductors 0.11V @ 10 feet
NUM_TURN_LEDS = (
    NUM_TURN_LEDS_1 +
    NUM_TURN_LEDS_2 +
    NUM_TURN_LEDS_3
)


RIGHT_MARKER_PIN = machine.Pin(23, machine.Pin.OUT)
RIGHT_TURN_PIN_3 = machine.Pin(22, machine.Pin.OUT)
RIGHT_TURN_PIN_2 = machine.Pin(21, machine.Pin.OUT)
RIGHT_TURN_PIN_1 = machine.Pin(19, machine.Pin.OUT)
RIGHT_FOG_PIN = machine.Pin(18, machine.Pin.OUT)

LEFT_FOG_PIN = machine.Pin(5, machine.Pin.OUT)
LEFT_TURN_PIN_1 = machine.Pin(17, machine.Pin.OUT)
LEFT_TURN_PIN_2 = machine.Pin(16, machine.Pin.OUT)
LEFT_TURN_PIN_3 = machine.Pin(4, machine.Pin.OUT)
LEFT_MARKER_PIN = machine.Pin(2, machine.Pin.OUT)


LEFT_PIN = machine.Pin(12, machine.Pin.IN, machine.Pin.PULL_DOWN)
RIGHT_PIN = machine.Pin(14, machine.Pin.IN, machine.Pin.PULL_DOWN)
MARKER_PIN = machine.Pin(27, machine.Pin.IN, machine.Pin.PULL_DOWN)
FOG_PIN = machine.Pin(26, machine.Pin.IN, machine.Pin.PULL_DOWN)
LOW_PIN = machine.Pin(25, machine.Pin.IN, machine.Pin.PULL_DOWN)
HIGH_PIN = machine.Pin(33, machine.Pin.IN, machine.Pin.PULL_DOWN)
RESET_PIN = machine.Pin(13, machine.Pin.IN, machine.Pin.PULL_UP)


def rgb_to_hsv(r, g, b, _):
    r /= 255.0
    g /= 255.0
    b /= 255.0
    maxc = max(r, g, b)
    minc = min(r, g, b)
    v = maxc

    if minc == maxc:
        h, s = 0.0, 0.0
    else:
        s = (maxc - minc) / maxc
        rc = (maxc - r) / (maxc - minc)
        gc = (maxc - g) / (maxc - minc)
        bc = (maxc - b) / (maxc - minc)

        if r == maxc:
            h = bc - gc
        elif g == maxc:
            h = 2.0 + rc-bc
        else:
            h = 4.0 + gc-rc

        h = (h / 6.0) % 1.0

    return h * 360.0, s * 100.0, v * 100.0


def hsv_to_rgb(h, s, v):
    h /= 360.0
    s /= 100.0
    v /= 100.0

    if s == 0.0:
        r = g = b = v
    else:
        k = int(h * 6.0)
        f = (h * 6.0) - k
        p = v * (1.0 - s)
        q = v * (1.0 - s * f)
        t = v * (1.0 - s * (1.0 - f))
        k %= 6

        if k == 0:
            r, g, b = v, t, p
        elif k == 1:
            r, g, b = q, v, p
        elif k == 2:
            r, g, b = p, v, t
        elif k == 3:
            r, g, b = p, q, v
        elif k == 4:
            r, g, b = t, p, v
        else:
            r, g, b = v, p, q

    return int(r * 255.0), int(g * 255.0), int(b * 255.0), 0


class Marker(object):

    def __init__(self, led_pin):
        self.np = neopixel.NeoPixel(led_pin, NUM_MARKER_LEDS, 4)
        self.write_pixels = False
        self._exit = False
        self._lock = _thread.allocate_lock()
        self._lock.acquire()
        self._t = _thread.start_new_thread(self._loop, ())

    def exit(self):
        if self._t is not None:
            self._exit = True
            self._lock.release()

    def write(self):
        self._lock.release()

    def _loop(self, *_, **__):
        for j in range(NUM_MARKER_LEDS):
            self.np[j] = BLACK

        self.np.write()
        
        while not self._exit:
            self._lock.acquire()
            self.np.write()
            self.write_pixels = False

        self._t = None

    def __setitem__(self, key, value):
        if value != self.np[key]:
            self.np[key] = value
            self.write_pixels = True

    def __getitem__(self, index):
        return self.np[index]

    def __len__(self):
        return NUM_MARKER_LEDS

    def pop(self, index):
        item = self.np[index]

        for j in range(index, NUM_MARKER_LEDS - 1):
            self.np[j] = self.np[j + 1]

        self.np[NUM_MARKER_LEDS - 1] = BLACK

        return item

    def insert(self, index, item):
        for j in range(index, NUM_MARKER_LEDS - 1):
            self.np[j + 1] = self.np[j]

        self.np[index] = item
        self.write_pixels = True

    def append(self, item):
        self.np[NUM_FOG_LEDS - 1] = item
        self.write_pixels = True


class Fog(object):
    def __init__(self, led_pin):
        self.np = neopixel.NeoPixel(led_pin, NUM_FOG_LEDS, 4)
        self.write_pixels = False
        self._exit = False
        self._lock = _thread.allocate_lock()
        self._lock.acquire()
        self._t = _thread.start_new_thread(self._loop, ())

    def exit(self):
        if self._t is not None:
            self._exit = True
            self._lock.release()

    def write(self):
        self._lock.release()

    def _loop(self, *_, **__):
        for j in range(NUM_FOG_LEDS):
            self.np[j] = BLACK

        self.np.write()

        while not self._exit:
            self._lock.acquire()
            self.np.write()
            self.write_pixels = False

        self._t = None

    def __setitem__(self, key, value):
        if value != self.np[key]:
            self.np[key] = value
            self.write_pixels = True

    def __getitem__(self, index):
        return self.np[index]

    def __len__(self):
        return NUM_FOG_LEDS

    def pop(self, index):

        item = self.np[index]

        for j in range(index, NUM_FOG_LEDS - 1):
            self.np[j] = self.np[j + 1]

        self.np[NUM_FOG_LEDS - 1] = BLACK
        return item

    def insert(self, index, item):

        for j in range(index, NUM_FOG_LEDS - 1):
            self.np[j + 1] = self.np[j]

        self.np[index] = item
        self.write_pixels = True

    def append(self, item):
        self.np[NUM_FOG_LEDS - 1] = item
        self.write_pixels = True


class Turn(object):

    def __init__(self, led_pin1, led_pin2, led_pin3):
        self.np1 = neopixel.NeoPixel(led_pin1, NUM_TURN_LEDS_1, 4)
        self.np2 = neopixel.NeoPixel(led_pin2, NUM_TURN_LEDS_2, 4)
        self.np3 = neopixel.NeoPixel(led_pin3, NUM_TURN_LEDS_3, 4)
        self.write_np1_pixels = False
        self.write_np2_pixels = False
        self.write_np3_pixels = False
        self._exit = False
        self._lock = _thread.allocate_lock()
        self._lock.acquire()
        self._t = _thread.start_new_thread(self._loop, ())

    def exit(self):
        if self._t is not None:
            self._exit = True
            self._lock.release()

    def write(self):
        self._lock.release()

    def _loop(self, *_, **__):
        for j in range(NUM_TURN_LEDS_1):
            self.np1[j] = BLACK

        self.np1.write()

        for j in range(NUM_TURN_LEDS_2):
            self.np2[j] = BLACK

        self.np2.write()

        for j in range(NUM_TURN_LEDS_3):
            self.np3[j] = BLACK

        self.np3.write()

        while not self._exit:
            self._lock.acquire()
            if self.write_np1_pixels:
                self.np1.write()
                self.write_np1_pixels = False

            if self.write_np2_pixels:
                self.np2.write()
                self.write_np2_pixels = False

            if self.write_np3_pixels:
                self.np3.write()
                self.write_np3_pixels = False

        self._t = None

    def __setitem__(self, key, value):
        if key < NUM_TURN_LEDS_1:
            if value != self.np1[key]:
                self.np1[key] = value
                self.write_np1_pixels = True

        elif key < NUM_TURN_LEDS_2:
            key -= NUM_TURN_LEDS_1

            if value != self.np2[key]:
                self.np2[key] = value
                self.write_np2_pixels = True
        elif key < NUM_TURN_LEDS_3:
            key -= NUM_TURN_LEDS_1 - NUM_TURN_LEDS_2

            if value != self.np3[key]:
                self.np3[key] = value
                self.write_np3_pixels = True
        else:
            raise IndexError(key)

    def __getitem__(self, index):
        if index < NUM_TURN_LEDS_1:
            return self.np1[index]
        elif index < NUM_TURN_LEDS_2:
            index -= NUM_TURN_LEDS_1
            return self.np2[index]
        elif index < NUM_TURN_LEDS_3:
            index -= NUM_TURN_LEDS_1 - NUM_TURN_LEDS_2
            return self.np3[index]

        raise IndexError(index)

    def __len__(self):
        return NUM_TURN_LEDS

    def pop(self, index):
        if index < NUM_TURN_LEDS_1:
            item = self.np1[index]
            self.write_np1_pixels = True
            self.write_np2_pixels = True
            self.write_np3_pixels = True

            for j in range(index, NUM_TURN_LEDS_1 - 1):
                self.np1[j] = self.np1[j + 1]

            self.np1[NUM_TURN_LEDS_1 - 1] = self.np2[0]

            for j in range(NUM_TURN_LEDS_2 - 1):
                self.np2[j] = self.np2[j + 1]

            self.np2[NUM_TURN_LEDS_2 - 1] = self.np3[0]

            for j in range(NUM_TURN_LEDS_3 - 1):
                self.np3[j] = self.np3[j + 1]

            self.np3[NUM_TURN_LEDS_3 - 1] = BLACK

        elif index < NUM_TURN_LEDS_2:
            index -= NUM_TURN_LEDS_1
            item = self.np2[index]

            self.write_np2_pixels = True
            self.write_np3_pixels = True

            for j in range(index, NUM_TURN_LEDS_2 - 1):
                self.np2[j] = self.np2[j + 1]

            self.np2[NUM_TURN_LEDS_2 - 1] = self.np3[0]

            for j in range(NUM_TURN_LEDS_3 - 1):
                self.np3[j] = self.np3[j + 1]

            self.np3[NUM_TURN_LEDS_3 - 1] = BLACK

        elif index < NUM_TURN_LEDS_3:
            index -= NUM_TURN_LEDS_1 - NUM_TURN_LEDS_2
            item = self.np3[index]

            self.write_np3_pixels = True

            for j in range(index, NUM_TURN_LEDS_3 - 1):
                self.np3[j] = self.np3[j + 1]

            self.np3[NUM_TURN_LEDS_3 - 1] = BLACK

        else:
            raise IndexError(index)

        return item

    def insert(self, index, item):
        if index < NUM_TURN_LEDS_1:
            self.write_np1_pixels = True
            self.write_np2_pixels = True
            self.write_np3_pixels = True

            for j in range(NUM_TURN_LEDS_3 - 1):
                self.np3[j + 1] = self.np3[j]

            self.np3[0] = self.np2[NUM_TURN_LEDS_2 - 1]

            for j in range(NUM_TURN_LEDS_2 - 1):
                self.np2[j + 1] = self.np2[j]

            self.np2[0] = self.np1[NUM_TURN_LEDS_1 - 1]

            for j in range(index, NUM_TURN_LEDS_1 - 1):
                self.np1[j + 1] = self.np1[j]

            self.np1[index] = item

        elif index < NUM_TURN_LEDS_2:
            index -= NUM_TURN_LEDS_1

            self.write_np2_pixels = True
            self.write_np3_pixels = True

            for j in range(NUM_TURN_LEDS_3 - 1):
                self.np3[j + 1] = self.np3[j]

            self.np3[0] = self.np2[NUM_TURN_LEDS_2 - 1]

            for j in range(index, NUM_TURN_LEDS_2 - 1):
                self.np2[j + 1] = self.np2[j]

            self.np2[index] = item

        elif index < NUM_TURN_LEDS_3:
            index -= NUM_TURN_LEDS_1 - NUM_TURN_LEDS_2

            self.write_np3_pixels = True

            for j in range(index, NUM_TURN_LEDS_3 - 1):
                self.np3[j + 1] = self.np3[j]

            self.np3[index] = item

        else:
            raise IndexError(index)

    def append(self, item):
        self.np3[NUM_TURN_LEDS_3 - 1] = item
        self.write_np3_pixels = True


class Timer(object):

    def __init__(self, duration):
        self.duration = duration
        self._start = utime.ticks_ms()
        self._t = None

    def reset(self):
        self._start = utime.ticks_ms()
        
    @property
    def is_expired(self):
        return utime.ticks_diff(utime.ticks_ms(), self._start) >= self.duration
    
    
class States(object):

    def __init__(self):
        self._left = False
        self._right = False
        self._marker = False
        self._fog = False
        self._high = False
        self._low = False
        self._center_fog = False
        
        self._left_changed = False
        self._right_changed = False
        self._center_fog_changed = False
        self._fog_changed = False
        self._marker_changed = False
        self._high_changed = False
        self._low_changed = False
        
        self._left_timer = None
        self._right_timer = None
        self._center_fog_timer = None
        
        self._lock = _thread.allocate_lock()
        self._lock.acquire()

    def wait(self):
        if self._left_timer is None and self._right_timer is None:
            self._lock.acquire()

    @property
    def left_changed(self):
        if self._left_timer is not None:
            if self._left_timer.is_elapsed:
                self._left = False
                self._left_changed = True
                self._left_timer = None
            
        changed = self._left_changed
        self._left_changed = False
        return changed

    @property
    def right_changed(self):
        if self._right_timer is not None:
            if self._right_timer.is_elapsed:
                self._right = False
                self._right_changed = True
                self._right_timer = None
                
        changed = self._right_changed
        self._right_changed = False
        return changed

    @property
    def center_fog_changed(self):
        changed = self._center_fog_changed
        self._center_fog_changed = False
        return changed

    @property
    def fog_changed(self):
        changed = self._fog_changed
        self._fog_changed = False
        return changed

    @property
    def marker_changed(self):
        changed = self._marker_changed
        self._marker_changed = False
        return changed

    @property
    def high_changed(self):
        changed = self._high_changed
        self._high_changed = False
        return changed

    @property
    def low_changed(self):
        changed = self._low_changed
        self._low_changed = False
        return changed

    @property
    def left(self):
        return self._left

    @left.setter
    def left(self, value):
        if value == self._left:
            return

        if self._left_timer is None:
            if value:
                self._left_timer = Timer(1.1)
            else:
                return
        else:
            self._left_timer.reset()
            return
           
        self._left = value
        self._left_changed = True
        
        if self._lock.locked():
            self._lock.release()

    @property
    def right(self):
        return self._right

    @right.setter
    def right(self, value):
        if value == self._right:
            return

        if self._right_timer is None:
            if value:
                self._right_timer = Timer(1.1)
            else:
                return
        else:
            self._right_timer.reset()
            return

        self._right = value
        self._right_changed = True

        if self._lock.locked():
            self._lock.release()

    @property
    def marker(self):
        return self._marker

    @marker.setter
    def marker(self, value):
        if value == self._marker:
            return

        self._marker = value
        self._marker_changed = True

        if self._lock.locked():
            self._lock.release()

    @property
    def fog(self):
        return self._fog

    @fog.setter
    def fog(self, value):
        if value == self._fog:
            return

        self._fog = value
        self._fog_changed = True

        if self._lock.locked():
            self._lock.release()
                
    @property
    def center_fog(self):
        return self._center_fog
    
    @center_fog.setter
    def center_fog(self, value):
        if value == self._fog:
            return

        self._fog = value
        self._center_fog_changed = True

        if self._lock.locked():
            self._lock.release()

    @property
    def high(self):
        return self._high

    @high.setter
    def high(self, value):
        if value == self._high:
            return

        self._high = value
        self._high_changed = True

        if self._lock.locked():
            self._lock.release()

    @property
    def low(self):
        return self._low

    @low.setter
    def low(self, value):
        if value == self._low:
            return

        self._low = value
        self._low_changed = True

        if self._lock.locked():
            self._lock.release()


states = States()


def left_callback(pin):
    states.left = bool(pin.value())
    

def right_callback(pin):
    states.right = bool(pin.value())


def marker_callback(pin):
    states.marker = bool(pin.value())


def fog_callback(pin):
    states.fog = bool(pin.value())


def high_callback(pin):
    states.high = bool(pin.value())


def low_callback(pin):
    states.low = bool(pin.value())


def reset_callback(_):
    pass


LEFT_PIN.irq(left_callback)
RIGHT_PIN.irq(right_callback)
MARKER_PIN.irq(marker_callback)
FOG_PIN.irq(fog_callback)
LOW_PIN.irq(low_callback)
HIGH_PIN.irq(high_callback)
RESET_PIN.irq(reset_callback)

AMBER = (255, 191, 0, 0)
DIM_AMBER = hsv_to_rgb(*rgb_to_hsv(*AMBER)[:-1] + (50.0,))
WHITE = (255, 255, 255, 255)
DIM_WHITE = (255, 255, 255, 0)
BLACK = (0, 0, 0, 0)

left_np = Turn(LEFT_TURN_PIN_1, LEFT_TURN_PIN_2, LEFT_TURN_PIN_3)
right_np = Turn(RIGHT_TURN_PIN_1, RIGHT_TURN_PIN_2, RIGHT_TURN_PIN_3)
left_marker_np = Marker(LEFT_MARKER_PIN)
right_marker_np = Marker(RIGHT_MARKER_PIN)
right_fog_np = Fog(RIGHT_FOG_PIN)
left_fog_np = Fog(LEFT_FOG_PIN)


def run_turn(pixels, turn_np, marker_np, fog_np):
    if pixels is None:
        pixels = [None] * NUM_TURN_LEDS

        if states.low or states.high:
            pixels[NUM_TURN_LEDS_1 + NUM_TURN_LEDS_2 - 1] = WHITE

        pixels[0] = AMBER

    else:
        if states.low or states.high:
            if left_turn_pixels[-1] is None:
                pixels.pop(NUM_TURN_LEDS - 1)
                pixels.insert(NUM_TURN_LEDS_1 + NUM_TURN_LEDS_2 - 1, WHITE)

            if left_turn_pixels[NUM_TURN_LEDS_1 + NUM_TURN_LEDS_2 - 2] is None:
                color = pixels[0]
                pixels.pop(NUM_TURN_LEDS_1 + NUM_TURN_LEDS_2 - 2)

                if color != BLACK:
                    hue, sat, val = rgb_to_hsv(*color)
                    val -= 100.0 / (NUM_TURN_LEDS_1 + NUM_TURN_LEDS_2)

                    if val < 0:
                        color = BLACK
                    else:
                        color = hsv_to_rgb(hue, sat, val)

                pixels.insert(0, color)

            else:
                pixels.insert(0, pixels.pop(NUM_TURN_LEDS_1 + NUM_TURN_LEDS_2 - 2))
        else:
            if pixels[-1] is None:
                color = pixels[0]
                pixels.pop(NUM_TURN_LEDS - 1)

                if color != BLACK:
                    hue, sat, val = rgb_to_hsv(*color)
                    val -= 100.0 / (NUM_TURN_LEDS_1 + NUM_TURN_LEDS_2)

                    if val < 0:
                        color = BLACK
                    else:
                        color = hsv_to_rgb(hue, sat, val)

                pixels.insert(0, color)

            else:
                pixels.insert(0, pixels.pop(NUM_TURN_LEDS_1 + NUM_TURN_LEDS_2 - 2))

        for color, i in enumerate(pixels):
            if color is None:
                continue

            turn_np[i] = color

        turn_np.write()

        if fog_np[NUM_FOG_LEDS - 1] != BLACK:
            fog_np.insert(0, BLACK)
            fog_np.write()
            
        if marker_np[NUM_MARKER_LEDS - 1] not in (BLACK, AMBER):
            marker_np.insert(0, AMBER)
        elif marker_np[NUM_MARKER_LEDS - 1] == BLACK and marker_np[0] == BLACK:
            marker_np.insert(0, AMBER)
        elif marker_np[NUM_MARKER_LEDS - 1] == AMBER and marker_np[0] == AMBER:
            marker_np.insert(0, BLACK)
        elif marker_np[0] == AMBER:
            marker_np.insert(0, AMBER)
        else:
            marker_np.insert(0, BLACK)
        marker_np.write()
        
    return pixels[:]


def stop_turn(turn_np, marker_np, fog_np):
    if states.fog:
        turn_color = WHITE
    else:
        turn_color = BLACK

    if states.marker:
        marker_color = DIM_AMBER
    else:
        marker_color = BLACK

    for i in range(NUM_TURN_LEDS):
        if i < NUM_FOG_LEDS and states.center_fog:
            fog_np[i] = WHITE
            fog_np.write()

        if i < NUM_MARKER_LEDS:
            marker_np[i] = marker_color
            marker_np.write()

        if i < 25:
            if states.fog and states.marker:
                turn_np[i] = AMBER
            elif states.fog:
                turn_np[i] = WHITE
            else:
                turn_np[i] = marker_color
        else:
            turn_np[i] = turn_color

        turn_np.write()


left_turn_pixels = None
right_turn_pixels = None


def change_center_fog(fog_np):
    if states.center_fog:
        color = WHITE
    else:
        color = BLACK

    if fog_np[0] != color:
        fog_np.pop(0)
        fog_np.ppend(color)
        fog_np.write()


def change_fog(turn_np):
    if states.marker:
        index = 26
    else:
        index = 0

    if states.fog:
        color = WHITE
    elif states.marker:
        color = DIM_AMBER
    else:
        color = BLACK

    if turn_np[index] != color:
        turn_np.pop(index)
        turn_np.insert(NUM_TURN_LEDS - 1, color)

        turn_np.write()


def change_marker(turn_np, marker_np):
    if states.marker:
        marker_color = DIM_AMBER
        if states.fog:
            color = AMBER
            index = 25
        else:
            color = DIM_AMBER
            index = NUM_TURN_LEDS - 1
    else:
        if states.fog:
            color = WHITE
            index = 25
        else:
            color = BLACK
            index = NUM_TURN_LEDS - 1

        marker_color = BLACK

    if turn_np[index] != color:
        turn_np.pop(0)
        turn_np.insert(index, color)
        turn_np.write()

    if marker_np[0] != marker_color:
        marker_np.pop(0)
        marker_np.append(marker_color)
        marker_np.write()


while True:
    states.wait()

    if states.left_changed and not states.left:
        stop_turn(left_np, left_marker_np, left_fog_np)
        left_turn_pixels = None

    if states.left:
        left_turn_pixels = run_turn(left_turn_pixels, left_np, left_marker_np, left_fog_np)
    else:
        change_center_fog(left_fog_np)
        change_fog(left_np)
        change_marker(left_np, left_marker_np)

    if states.right_changed and not states.right:
        stop_turn(right_np, right_marker_np, right_fog_np)
        right_turn_pixels = None

    if states.right:
        right_turn_pixels = run_turn(right_turn_pixels, right_np, right_marker_np, right_fog_np)
    else:
        change_center_fog(right_fog_np)
        change_fog(right_np)
        change_marker(right_np, right_marker_np)
